package solutions.bitbadger.documents.common

/**
 * Information required to generate a JSON field comparison
 */
interface Comparison<T> {

    /** The operation for the field comparison */
    val op: Op

    /** The value against which the comparison will be made */
    val value: T

    /** Whether the value should be considered numeric */
    val isNumeric: Boolean
}

/**
 * Function to determine if a value is numeric
 *
 * @param it The value in question
 * @return True if it is a numeric type, false if not
 */
private fun <T> isNumeric(it: T) =
    it is Byte || it is Short || it is Int || it is Long

/**
 * A single-value comparison against a field in a JSON document
 */
class ComparisonSingle<T>(override val op: Op, override val value: T) : Comparison<T> {

    init {
        when (op) {
            Op.BETWEEN, Op.IN, Op.IN_ARRAY ->
                throw DocumentException("Cannot use single comparison for multiple values")
            else -> { }
        }
    }

    override val isNumeric = isNumeric(value)

    override fun toString() =
        "$op $value"
}

/**
 * A range comparison against a field in a JSON document
 */
class ComparisonBetween<T>(override val value: Pair<T, T>) : Comparison<Pair<T, T>> {
    override val op        = Op.BETWEEN
    override val isNumeric = isNumeric(value.first)
}

/**
 * A check within a collection of values
 */
class ComparisonIn<T>(override val value: Collection<T>) : Comparison<Collection<T>> {
    override val op        = Op.IN
    override val isNumeric = !value.isEmpty() && isNumeric(value.elementAt(0))
}

/**
 * A check within a collection of values against an array in a document
 */
class ComparisonInArray<T>(override val value: Pair<String, Collection<T>>) : Comparison<Pair<String, Collection<T>>> {
    override val op        = Op.IN_ARRAY
    override val isNumeric = false
}
