package solutions.bitbadger.documents.common.query

import solutions.bitbadger.documents.common.*

/**
 * Functions to create queries to define tables and indexes
 */
object Definition {

    /**
     * SQL statement to create a document table
     *
     * @param tableName The name of the table to create (may include schema)
     * @param dataType The type of data for the column (`JSON`, `JSONB`, etc.)
     * @return A query to create a document table
     */
    fun ensureTableFor(tableName: String, dataType: String) =
        "CREATE TABLE IF NOT EXISTS $tableName (data $dataType NOT NULL)"

    /**
     * SQL statement to create a document table in the current dialect
     *
     * @param tableName The name of the table to create (may include schema)
     * @param dialect The dialect to generate (optional, used in place of current)
     * @return A query to create a document table
     */
    fun ensureTable(tableName: String, dialect: Dialect? = null) =
        when (dialect ?: Configuration.dialect("create table creation query")) {
            Dialect.POSTGRESQL -> ensureTableFor(tableName, "JSONB")
            Dialect.SQLITE     -> ensureTableFor(tableName, "TEXT")
        }

    /**
     * Split a schema and table name
     *
     * @param tableName The name of the table, possibly with a schema
     * @return A pair with the first item as the schema and the second as the table name
     */
    private fun splitSchemaAndTable(tableName: String) =
        tableName.split('.').let { if (it.size == 1) Pair("", tableName) else Pair(it[0], it[1]) }

    /**
     * SQL statement to create an index on one or more fields in a JSON document
     *
     * @param tableName The table on which an index should be created (may include schema)
     * @param indexName The name of the index to be created
     * @param fields One or more fields to include in the index
     * @param dialect The SQL dialect to use when creating this index (optional, used in place of current)
     * @return A query to create the field index
     */
    fun ensureIndexOn(
        tableName: String,
        indexName: String,
        fields: Collection<String>,
        dialect: Dialect? = null
    ): String {
        val (_, tbl) = splitSchemaAndTable(tableName)
        val mode = dialect ?: Configuration.dialect("create index $tbl.$indexName")
        val jsonFields = fields.joinToString(", ") {
            val parts = it.split(' ')
            val direction = if (parts.size > 1) " ${parts[1]}" else ""
            "(" + Field.nameToPath(parts[0], mode, FieldFormat.SQL) + ")$direction"
        }
        return "CREATE INDEX IF NOT EXISTS idx_${tbl}_$indexName ON $tableName ($jsonFields)"
    }

    /**
     * SQL statement to create a key index for a document table
     *
     * @param tableName The table on which a key index should be created (may include schema)
     * @param dialect The SQL dialect to use when creating this index (optional, used in place of current)
     * @return A query to create the key index
     */
    fun ensureKey(tableName: String, dialect: Dialect? = null) =
        ensureIndexOn(tableName, "key", listOf(Configuration.idField), dialect).replace("INDEX", "UNIQUE INDEX")

    /**
     * Create a document-wide index on a table (PostgreSQL only)
     *
     * @param tableName The name of the table on which the document index should be created
     * @param indexType The type of index to be created
     * @return The SQL statement to create an index on JSON documents in the specified table
     * @throws DocumentException If the database mode is not PostgreSQL
     */
    fun ensureDocumentIndexOn(tableName: String, indexType: DocumentIndex): String {
        if (Configuration.dialect("create document index query") != Dialect.POSTGRESQL) {
            throw DocumentException("'Document indexes are only supported on PostgreSQL")
        }
        val (_, tbl) = splitSchemaAndTable(tableName)
        return "CREATE INDEX IF NOT EXISTS idx_${tbl}_document ON $tableName USING GIN (data${indexType.sql})"
    }
}
