package solutions.bitbadger.documents.common.query

import solutions.bitbadger.documents.common.AutoId
import solutions.bitbadger.documents.common.Configuration
import solutions.bitbadger.documents.common.Dialect

/**
 * Functions for document-level operations
 */
object Document {

    /**
     * Query to insert a document
     *
     * @param tableName The table into which to insert (may include schema)
     * @return A query to insert a document
     */
    fun insert(tableName: String, autoId: AutoId? = null): String {
        val id = Configuration.idField
        val values = when (Configuration.dialect("create INSERT statement")) {
            Dialect.POSTGRESQL -> when (autoId ?: AutoId.DISABLED) {
                AutoId.DISABLED      -> ":data"
                AutoId.NUMBER        -> ":data::jsonb || ('{\"$id\":' || " +
                        "(SELECT COALESCE(MAX((data->>'$id')::numeric), 0) + 1 " +
                        "FROM $tableName) || '}')::jsonb"
                AutoId.UUID          -> ":data::jsonb || '{\"$id\":\"${AutoId.generateUUID()}\"}'"
                AutoId.RANDOM_STRING -> ":data::jsonb || '{\"$id\":\"${AutoId.generateRandomString()}\"}'"
            }
            Dialect.SQLITE -> when (autoId ?: AutoId.DISABLED) {
                AutoId.DISABLED      -> ":data"
                AutoId.NUMBER        -> "json_set(:data, '$.$id', " +
                        "(SELECT coalesce(max(data->>'$id'), 0) + 1 FROM $tableName))"
                AutoId.UUID          -> "json_set(:data, '$.$id', '${AutoId.generateUUID()}')"
                AutoId.RANDOM_STRING -> "json_set(:data, '$.$id', '${AutoId.generateRandomString()}')"
            }
        }
        return "INSERT INTO $tableName VALUES ($values)"
    }

    /**
     * Query to save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")
     *
     * @param tableName The table into which to save (may include schema)
     * @return A query to save a document
     */
    fun save(tableName: String) =
        insert(tableName, AutoId.DISABLED) +
                " ON CONFLICT ((data->>'${Configuration.idField}')) DO UPDATE SET data = EXCLUDED.data"

    /**
     * Query to update (replace) a document (this query has no `WHERE` clause)
     *
     * @param tableName The table in which documents should be replaced (may include schema)
     * @return A query to update documents
     */
    fun update(tableName: String) =
        "UPDATE $tableName SET data = :data"
}
