package solutions.bitbadger.documents.common.java;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import solutions.bitbadger.documents.common.AutoId;
import solutions.bitbadger.documents.common.DocumentException;
import solutions.bitbadger.documents.java.java.testDocs.*;

import static org.junit.jupiter.api.Assertions.*;

/**
 * Unit tests for the `AutoId` enum
 */
@DisplayName("Java | Common | AutoId")
final public class AutoIdTest {

    @Test
    @DisplayName("Generates a UUID string")
    public void generateUUID() {
        assertEquals(32, AutoId.generateUUID().length(), "The UUID should have been a 32-character string");
    }

    @Test
    @DisplayName("Generates a random hex character string of an even length")
    public void generateRandomStringEven() {
        final String result = AutoId.generateRandomString(8);
        assertEquals(8, result.length(), "There should have been 8 characters in " + result);
    }

    @Test
    @DisplayName("Generates a random hex character string of an odd length")
    public void generateRandomStringOdd() {
        final String result = AutoId.generateRandomString(11);
        assertEquals(11, result.length(), "There should have been 11 characters in " + result);
    }

    @Test
    @DisplayName("Generates different random hex character strings")
    public void generateRandomStringIsRandom() {
        final String result1 = AutoId.generateRandomString(16);
        final String result2 = AutoId.generateRandomString(16);
        assertNotEquals(result1, result2, "There should have been 2 different strings generated");
    }

    @Test
    @DisplayName("needsAutoId fails for null document")
    public void needsAutoIdFailsForNullDocument() {
        assertThrows(DocumentException.class, () -> AutoId.needsAutoId(AutoId.DISABLED, null, "id"));
    }

    @Test
    @DisplayName("needsAutoId fails for missing ID property")
    public void needsAutoIdFailsForMissingId() {
        assertThrows(DocumentException.class, () -> AutoId.needsAutoId(AutoId.UUID, new IntIdClass(0), "Id"));
    }

    @Test
    @DisplayName("needsAutoId returns false if disabled")
    public void needsAutoIdFalseIfDisabled() {
        try {
            assertFalse(AutoId.needsAutoId(AutoId.DISABLED, "", ""), "Disabled Auto ID should always return false");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId returns true for Number strategy and byte ID of 0")
    public void needsAutoIdTrueForByteWithZero() {
        try {
            assertTrue(AutoId.needsAutoId(AutoId.NUMBER, new ByteIdClass((byte) 0), "id"),
                    "Number Auto ID with 0 should return true");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId returns false for Number strategy and byte ID of non-0")
    public void needsAutoIdFalseForByteWithNonZero() {
        try {
            assertFalse(AutoId.needsAutoId(AutoId.NUMBER, new ByteIdClass((byte) 77), "id"),
                    "Number Auto ID with 77 should return false");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId returns true for Number strategy and short ID of 0")
    public void needsAutoIdTrueForShortWithZero() {
        try {
            assertTrue(AutoId.needsAutoId(AutoId.NUMBER, new ShortIdClass((short) 0), "id"),
                    "Number Auto ID with 0 should return true");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId returns false for Number strategy and short ID of non-0")
    public void needsAutoIdFalseForShortWithNonZero() {
        try {
            assertFalse(AutoId.needsAutoId(AutoId.NUMBER, new ShortIdClass((short) 31), "id"),
                    "Number Auto ID with 31 should return false");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId returns true for Number strategy and int ID of 0")
    public void needsAutoIdTrueForIntWithZero() {
        try {
            assertTrue(AutoId.needsAutoId(AutoId.NUMBER, new IntIdClass(0), "id"),
                    "Number Auto ID with 0 should return true");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId returns false for Number strategy and int ID of non-0")
    public void needsAutoIdFalseForIntWithNonZero() {
        try {
            assertFalse(AutoId.needsAutoId(AutoId.NUMBER, new IntIdClass(6), "id"),
                    "Number Auto ID with 6 should return false");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId returns true for Number strategy and long ID of 0")
    public void needsAutoIdTrueForLongWithZero() {
        try {
            assertTrue(AutoId.needsAutoId(AutoId.NUMBER, new LongIdClass(0L), "id"),
                    "Number Auto ID with 0 should return true");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId returns false for Number strategy and long ID of non-0")
    public void needsAutoIdFalseForLongWithNonZero() {
        try {
        assertFalse(AutoId.needsAutoId(AutoId.NUMBER, new LongIdClass(2L), "id"),
                "Number Auto ID with 2 should return false");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId fails for Number strategy and non-number ID")
    public void needsAutoIdFailsForNumberWithStringId() {
        assertThrows(DocumentException.class, () -> AutoId.needsAutoId(AutoId.NUMBER, new StringIdClass(""), "id"));
    }

    @Test
    @DisplayName("needsAutoId returns true for UUID strategy and blank ID")
    public void needsAutoIdTrueForUUIDWithBlank() {
        try {
            assertTrue(AutoId.needsAutoId(AutoId.UUID, new StringIdClass(""), "id"),
                    "UUID Auto ID with blank should return true");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId returns false for UUID strategy and non-blank ID")
    public void needsAutoIdFalseForUUIDNotBlank() {
        try {
            assertFalse(AutoId.needsAutoId(AutoId.UUID, new StringIdClass("howdy"), "id"),
                    "UUID Auto ID with non-blank should return false");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId fails for UUID strategy and non-string ID")
    public void needsAutoIdFailsForUUIDNonString() {
        assertThrows(DocumentException.class, () -> AutoId.needsAutoId(AutoId.UUID, new IntIdClass(5), "id"));
    }

    @Test
    @DisplayName("needsAutoId returns true for Random String strategy and blank ID")
    public void needsAutoIdTrueForRandomWithBlank() {
        try {
            assertTrue(AutoId.needsAutoId(AutoId.RANDOM_STRING, new StringIdClass(""), "id"),
                    "Random String Auto ID with blank should return true");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId returns false for Random String strategy and non-blank ID")
    public void needsAutoIdFalseForRandomNotBlank() {
        try {
            assertFalse(AutoId.needsAutoId(AutoId.RANDOM_STRING, new StringIdClass("full"), "id"),
                    "Random String Auto ID with non-blank should return false");
        } catch (DocumentException ex) {
            fail(ex);
        }
    }

    @Test
    @DisplayName("needsAutoId fails for Random String strategy and non-string ID")
    public void needsAutoIdFailsForRandomNonString() {
        assertThrows(DocumentException.class,
                () -> AutoId.needsAutoId(AutoId.RANDOM_STRING, new ShortIdClass((short) 55), "id"));
    }
}
