package solutions.bitbadger.documents.common.java;

import kotlin.Pair;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import solutions.bitbadger.documents.common.*;

import java.util.Collection;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

/**
 * Unit tests for the `Field` class
 */
@DisplayName("Java | Common | Field")
final public class FieldTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    public void cleanUp() {
        Configuration.setConnectionString(null);
    }

    // ~~~ INSTANCE METHODS ~~~

    @Test
    @DisplayName("withParameterName fails for invalid name")
    public void withParamNameFails() {
        assertThrows(DocumentException.class, () -> Field.equal("it", "").withParameterName("2424"));
    }

    @Test
    @DisplayName("withParameterName works with colon prefix")
    public void withParamNameColon() {
        Field<String> field = Field.equal("abc", "22").withQualifier("me");
        Field<String> withParam = field.withParameterName(":test");
        assertNotSame(field, withParam, "A new Field instance should have been created");
        assertEquals(field.getName(), withParam.getName(), "Name should have been preserved");
        assertEquals(field.getComparison(), withParam.getComparison(), "Comparison should have been preserved");
        assertEquals(":test", withParam.getParameterName(), "Parameter name not set correctly");
        assertEquals(field.getQualifier(), withParam.getQualifier(), "Qualifier should have been preserved");
    }

    @Test
    @DisplayName("withParameterName works with at-sign prefix")
    public void withParamNameAtSign() {
        Field<String> field = Field.equal("def", "44");
        Field<String> withParam = field.withParameterName("@unit");
        assertNotSame(field, withParam, "A new Field instance should have been created");
        assertEquals(field.getName(), withParam.getName(), "Name should have been preserved");
        assertEquals(field.getComparison(), withParam.getComparison(), "Comparison should have been preserved");
        assertEquals("@unit", withParam.getParameterName(), "Parameter name not set correctly");
        assertEquals(field.getQualifier(), withParam.getQualifier(), "Qualifier should have been preserved");
    }

    @Test
    @DisplayName("withQualifier sets qualifier correctly")
    public void withQualifier() {
        Field<String> field = Field.equal("j", "k");
        Field<String> withQual = field.withQualifier("test");
        assertNotSame(field, withQual, "A new Field instance should have been created");
        assertEquals(field.getName(), withQual.getName(), "Name should have been preserved");
        assertEquals(field.getComparison(), withQual.getComparison(), "Comparison should have been preserved");
        assertEquals(field.getParameterName(), withQual.getParameterName(),
                "Parameter Name should have been preserved");
        assertEquals("test", withQual.getQualifier(), "Qualifier not set correctly");
    }

    @Test
    @DisplayName("path generates for simple unqualified PostgreSQL field")
    public void pathPostgresSimpleUnqualified() {
        assertEquals("data->>'SomethingCool'",
                Field.greaterOrEqual("SomethingCool", 18).path(Dialect.POSTGRESQL, FieldFormat.SQL),
                "Path not correct");
    }

    @Test
    @DisplayName("path generates for simple qualified PostgreSQL field")
    public void pathPostgresSimpleQualified() {
        assertEquals("this.data->>'SomethingElse'",
                Field.less("SomethingElse", 9).withQualifier("this").path(Dialect.POSTGRESQL, FieldFormat.SQL),
                "Path not correct");
    }

    @Test
    @DisplayName("path generates for nested unqualified PostgreSQL field")
    public void pathPostgresNestedUnqualified() {
        assertEquals("data#>>'{My,Nested,Field}'",
                Field.equal("My.Nested.Field", "howdy").path(Dialect.POSTGRESQL, FieldFormat.SQL), "Path not correct");
    }

    @Test
    @DisplayName("path generates for nested qualified PostgreSQL field")
    public void pathPostgresNestedQualified() {
        assertEquals("bird.data#>>'{Nest,Away}'",
                Field.equal("Nest.Away", "doc").withQualifier("bird").path(Dialect.POSTGRESQL, FieldFormat.SQL),
                "Path not correct");
    }

    @Test
    @DisplayName("path generates for simple unqualified SQLite field")
    public void pathSQLiteSimpleUnqualified() {
        assertEquals("data->>'SomethingCool'",
                Field.greaterOrEqual("SomethingCool", 18).path(Dialect.SQLITE, FieldFormat.SQL), "Path not correct");
    }

    @Test
    @DisplayName("path generates for simple qualified SQLite field")
    public void pathSQLiteSimpleQualified() {
        assertEquals("this.data->>'SomethingElse'",
                Field.less("SomethingElse", 9).withQualifier("this").path(Dialect.SQLITE, FieldFormat.SQL),
                "Path not correct");
    }

    @Test
    @DisplayName("path generates for nested unqualified SQLite field")
    public void pathSQLiteNestedUnqualified() {
        assertEquals("data->'My'->'Nested'->>'Field'",
                Field.equal("My.Nested.Field", "howdy").path(Dialect.SQLITE, FieldFormat.SQL), "Path not correct");
    }

    @Test
    @DisplayName("path generates for nested qualified SQLite field")
    public void pathSQLiteNestedQualified() {
        assertEquals("bird.data->'Nest'->>'Away'",
                Field.equal("Nest.Away", "doc").withQualifier("bird").path(Dialect.SQLITE, FieldFormat.SQL),
                "Path not correct");
    }

    @Test
    @DisplayName("toWhere generates for exists w/o qualifier (PostgreSQL)")
    public void toWhereExistsNoQualPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("data->>'that_field' IS NOT NULL", Field.exists("that_field").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for exists w/o qualifier (SQLite)")
    public void toWhereExistsNoQualSQLite() {
        Configuration.setConnectionString(":sqlite:");
        assertEquals("data->>'that_field' IS NOT NULL", Field.exists("that_field").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for not-exists w/o qualifier (PostgreSQL)")
    public void toWhereNotExistsNoQualPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("data->>'a_field' IS NULL", Field.notExists("a_field").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for not-exists w/o qualifier (SQLite)")
    public void toWhereNotExistsNoQualSQLite() {
        Configuration.setConnectionString(":sqlite:");
        assertEquals("data->>'a_field' IS NULL", Field.notExists("a_field").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for BETWEEN w/o qualifier, numeric range (PostgreSQL)")
    public void toWhereBetweenNoQualNumericPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("(data->>'age')::numeric BETWEEN @agemin AND @agemax",
                Field.between("age", 13, 17, "@age").toWhere(), "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for BETWEEN w/o qualifier, alphanumeric range (PostgreSQL)")
    public void toWhereBetweenNoQualAlphaPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("data->>'city' BETWEEN :citymin AND :citymax",
                Field.between("city", "Atlanta", "Chicago", ":city").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for BETWEEN w/o qualifier (SQLite)")
    public void toWhereBetweenNoQualSQLite() {
        Configuration.setConnectionString(":sqlite:");
        assertEquals("data->>'age' BETWEEN @agemin AND @agemax", Field.between("age", 13, 17, "@age").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for BETWEEN w/ qualifier, numeric range (PostgreSQL)")
    public void toWhereBetweenQualNumericPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("(test.data->>'age')::numeric BETWEEN @agemin AND @agemax",
                Field.between("age", 13, 17, "@age").withQualifier("test").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for BETWEEN w/ qualifier, alphanumeric range (PostgreSQL)")
    public void toWhereBetweenQualAlphaPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("unit.data->>'city' BETWEEN :citymin AND :citymax",
                Field.between("city", "Atlanta", "Chicago", ":city").withQualifier("unit").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for BETWEEN w/ qualifier (SQLite)")
    public void toWhereBetweenQualSQLite() {
        Configuration.setConnectionString(":sqlite:");
        assertEquals("my.data->>'age' BETWEEN @agemin AND @agemax",
                Field.between("age", 13, 17, "@age").withQualifier("my").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for IN/any, numeric values (PostgreSQL)")
    public void toWhereAnyNumericPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("(data->>'even')::numeric IN (:nbr_0, :nbr_1, :nbr_2)",
                Field.any("even", List.of(2, 4, 6), ":nbr").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for IN/any, alphanumeric values (PostgreSQL)")
    public void toWhereAnyAlphaPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("data->>'test' IN (:city_0, :city_1)",
                Field.any("test", List.of("Atlanta", "Chicago"), ":city").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for IN/any (SQLite)")
    public void toWhereAnySQLite() {
        Configuration.setConnectionString(":sqlite:");
        assertEquals("data->>'test' IN (:city_0, :city_1)",
                Field.any("test", List.of("Atlanta", "Chicago"), ":city").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for inArray (PostgreSQL)")
    public void toWhereInArrayPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("data->'even' ??| ARRAY[:it_0, :it_1, :it_2, :it_3]",
                Field.inArray("even", "tbl", List.of(2, 4, 6, 8), ":it").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for inArray (SQLite)")
    public void toWhereInArraySQLite() {
        Configuration.setConnectionString(":sqlite:");
        assertEquals("EXISTS (SELECT 1 FROM json_each(tbl.data, '$.test') WHERE value IN (:city_0, :city_1))",
                Field.inArray("test", "tbl", List.of("Atlanta", "Chicago"), ":city").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for others w/o qualifier (PostgreSQL)")
    public void toWhereOtherNoQualPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("data->>'some_field' = :value", Field.equal("some_field", "", ":value").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates for others w/o qualifier (SQLite)")
    public void toWhereOtherNoQualSQLite() {
        Configuration.setConnectionString(":sqlite:");
        assertEquals("data->>'some_field' = :value", Field.equal("some_field", "", ":value").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates no-parameter w/ qualifier (PostgreSQL)")
    public void toWhereNoParamWithQualPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("test.data->>'no_field' IS NOT NULL", Field.exists("no_field").withQualifier("test").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates no-parameter w/ qualifier (SQLite)")
    public void toWhereNoParamWithQualSQLite() {
        Configuration.setConnectionString(":sqlite:");
        assertEquals("test.data->>'no_field' IS NOT NULL", Field.exists("no_field").withQualifier("test").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates parameter w/ qualifier (PostgreSQL)")
    public void toWhereParamWithQualPostgres() {
        Configuration.setConnectionString(":postgresql:");
        assertEquals("(q.data->>'le_field')::numeric <= :it",
                Field.lessOrEqual("le_field", 18, ":it").withQualifier("q").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    @Test
    @DisplayName("toWhere generates parameter w/ qualifier (SQLite)")
    public void toWhereParamWithQualSQLite() {
        Configuration.setConnectionString(":sqlite:");
        assertEquals("q.data->>'le_field' <= :it",
                Field.lessOrEqual("le_field", 18, ":it").withQualifier("q").toWhere(),
                "Field WHERE clause not generated correctly");
    }

    // ~~~ STATIC TESTS ~~~

    @Test
    @DisplayName("equal constructs a field w/o parameter name")
    public void equalCtor() {
        Field<Integer> field = Field.equal("Test", 14);
        assertEquals("Test", field.getName(), "Field name not filled correctly");
        assertEquals(Op.EQUAL, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals(14, field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("equal constructs a field w/ parameter name")
    public void equalParameterCtor() {
        Field<Integer> field = Field.equal("Test", 14, ":w");
        assertEquals("Test", field.getName(), "Field name not filled correctly");
        assertEquals(Op.EQUAL, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals(14, field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertEquals(":w", field.getParameterName(), "Field parameter name not filled correctly");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("greater constructs a field w/o parameter name")
    public void greaterCtor() {
        Field<String> field = Field.greater("Great", "night");
        assertEquals("Great", field.getName(), "Field name not filled correctly");
        assertEquals(Op.GREATER, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals("night", field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("greater constructs a field w/ parameter name")
    public void greaterParameterCtor() {
        Field<String> field = Field.greater("Great", "night", ":yeah");
        assertEquals("Great", field.getName(), "Field name not filled correctly");
        assertEquals(Op.GREATER, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals("night", field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertEquals(":yeah", field.getParameterName(), "Field parameter name not filled correctly");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("greaterOrEqual constructs a field w/o parameter name")
    public void greaterOrEqualCtor() {
        Field<Long> field = Field.greaterOrEqual("Nice", 88L);
        assertEquals("Nice", field.getName(), "Field name not filled correctly");
        assertEquals(Op.GREATER_OR_EQUAL, field.getComparison().getOp(),
                "Field comparison operation not filled correctly");
        assertEquals(88L, field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("greaterOrEqual constructs a field w/ parameter name")
    public void greaterOrEqualParameterCtor() {
        Field<Long> field = Field.greaterOrEqual("Nice", 88L, ":nice");
        assertEquals("Nice", field.getName(), "Field name not filled correctly");
        assertEquals(Op.GREATER_OR_EQUAL, field.getComparison().getOp(),
                "Field comparison operation not filled correctly");
        assertEquals(88L, field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertEquals(":nice", field.getParameterName(), "Field parameter name not filled correctly");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("less constructs a field w/o parameter name")
    public void lessCtor() {
        Field<String> field = Field.less("Lesser", "seven");
        assertEquals("Lesser", field.getName(), "Field name not filled correctly");
        assertEquals(Op.LESS, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals("seven", field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("less constructs a field w/ parameter name")
    public void lessParameterCtor() {
        Field<String> field = Field.less("Lesser", "seven", ":max");
        assertEquals("Lesser", field.getName(), "Field name not filled correctly");
        assertEquals(Op.LESS, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals("seven", field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertEquals(":max", field.getParameterName(), "Field parameter name not filled correctly");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("lessOrEqual constructs a field w/o parameter name")
    public void lessOrEqualCtor() {
        Field<String> field = Field.lessOrEqual("Nobody", "KNOWS");
        assertEquals("Nobody", field.getName(), "Field name not filled correctly");
        assertEquals(Op.LESS_OR_EQUAL, field.getComparison().getOp(),
                "Field comparison operation not filled correctly");
        assertEquals("KNOWS", field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("lessOrEqual constructs a field w/ parameter name")
    public void lessOrEqualParameterCtor() {
        Field<String> field = Field.lessOrEqual("Nobody", "KNOWS", ":nope");
        assertEquals("Nobody", field.getName(), "Field name not filled correctly");
        assertEquals(Op.LESS_OR_EQUAL, field.getComparison().getOp(),
                "Field comparison operation not filled correctly");
        assertEquals("KNOWS", field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertEquals(":nope", field.getParameterName(), "Field parameter name not filled correctly");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("notEqual constructs a field w/o parameter name")
    public void notEqualCtor() {
        Field<String> field = Field.notEqual("Park", "here");
        assertEquals("Park", field.getName(), "Field name not filled correctly");
        assertEquals(Op.NOT_EQUAL, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals("here", field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("notEqual constructs a field w/ parameter name")
    public void notEqualParameterCtor() {
        Field<String> field = Field.notEqual("Park", "here", ":now");
        assertEquals("Park", field.getName(), "Field name not filled correctly");
        assertEquals(Op.NOT_EQUAL, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals("here", field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertEquals(":now", field.getParameterName(), "Field parameter name not filled correctly");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("between constructs a field w/o parameter name")
    public void betweenCtor() {
        Field<Pair<Integer, Integer>> field = Field.between("Age", 18, 49);
        assertEquals("Age", field.getName(), "Field name not filled correctly");
        assertEquals(Op.BETWEEN, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals(18, field.getComparison().getValue().getFirst(),
                "Field comparison min value not filled correctly");
        assertEquals(49, field.getComparison().getValue().getSecond(),
                "Field comparison max value not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("between constructs a field w/ parameter name")
    public void betweenParameterCtor() {
        Field<Pair<Integer, Integer>> field = Field.between("Age", 18, 49, ":limit");
        assertEquals("Age", field.getName(), "Field name not filled correctly");
        assertEquals(Op.BETWEEN, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals(18, field.getComparison().getValue().getFirst(),
                "Field comparison min value not filled correctly");
        assertEquals(49, field.getComparison().getValue().getSecond(),
                "Field comparison max value not filled correctly");
        assertEquals(":limit", field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("any constructs a field w/o parameter name")
    public void anyCtor() {
        Field<Collection<Integer>> field = Field.any("Here", List.of(8, 16, 32));
        assertEquals("Here", field.getName(), "Field name not filled correctly");
        assertEquals(Op.IN, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals(List.of(8, 16, 32), field.getComparison().getValue(),
                "Field comparison value not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("any constructs a field w/ parameter name")
    public void anyParameterCtor() {
        Field<Collection<Integer>> field = Field.any("Here", List.of(8, 16, 32), ":list");
        assertEquals("Here", field.getName(), "Field name not filled correctly");
        assertEquals(Op.IN, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals(List.of(8, 16, 32), field.getComparison().getValue(),
                "Field comparison value not filled correctly");
        assertEquals(":list", field.getParameterName(), "Field parameter name not filled correctly");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("inArray constructs a field w/o parameter name")
    public void inArrayCtor() {
        Field<Pair<String, Collection<String>>> field = Field.inArray("ArrayField", "table", List.of("z"));
        assertEquals("ArrayField", field.getName(), "Field name not filled correctly");
        assertEquals(Op.IN_ARRAY, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals("table", field.getComparison().getValue().getFirst(),
                "Field comparison table not filled correctly");
        assertEquals(List.of("z"), field.getComparison().getValue().getSecond(),
                "Field comparison values not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("inArray constructs a field w/ parameter name")
    public void inArrayParameterCtor() {
        Field<Pair<String, Collection<String>>> field = Field.inArray("ArrayField", "table", List.of("z"), ":a");
        assertEquals("ArrayField", field.getName(), "Field name not filled correctly");
        assertEquals(Op.IN_ARRAY, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals("table", field.getComparison().getValue().getFirst(),
                "Field comparison table not filled correctly");
        assertEquals(List.of("z"), field.getComparison().getValue().getSecond(),
                "Field comparison values not filled correctly");
        assertEquals(":a", field.getParameterName(), "Field parameter name not filled correctly");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("exists constructs a field")
    public void existsCtor() {
        Field<String> field = Field.exists("Groovy");
        assertEquals("Groovy", field.getName(), "Field name not filled correctly");
        assertEquals(Op.EXISTS, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals("", field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("notExists constructs a field")
    public void notExistsCtor() {
        Field<String> field = Field.notExists("Groovy");
        assertEquals("Groovy", field.getName(), "Field name not filled correctly");
        assertEquals(Op.NOT_EXISTS, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals("", field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("named constructs a field")
    public void namedCtor() {
        Field<String> field = Field.named("Tacos");
        assertEquals("Tacos", field.getName(), "Field name not filled correctly");
        assertEquals(Op.EQUAL, field.getComparison().getOp(), "Field comparison operation not filled correctly");
        assertEquals("", field.getComparison().getValue(), "Field comparison value not filled correctly");
        assertNull(field.getParameterName(), "The parameter name should have been null");
        assertNull(field.getQualifier(), "The qualifier should have been null");
    }

    @Test
    @DisplayName("static constructors fail for invalid parameter name")
    public void staticCtorsFailOnParamName() {
        assertThrows(DocumentException.class, () -> Field.equal("a", "b", "that ain't it, Jack..."));
    }

    @Test
    @DisplayName("nameToPath creates a simple PostgreSQL SQL name")
    public void nameToPathPostgresSimpleSQL() {
        assertEquals("data->>'Simple'", Field.nameToPath("Simple", Dialect.POSTGRESQL, FieldFormat.SQL),
                "Path not constructed correctly");
    }

    @Test
    @DisplayName("nameToPath creates a simple SQLite SQL name")
    public void nameToPathSQLiteSimpleSQL() {
        assertEquals("data->>'Simple'", Field.nameToPath("Simple", Dialect.SQLITE, FieldFormat.SQL),
                "Path not constructed correctly");
    }

    @Test
    @DisplayName("nameToPath creates a nested PostgreSQL SQL name")
    public void nameToPathPostgresNestedSQL() {
        assertEquals("data#>>'{A,Long,Path,to,the,Property}'",
                Field.nameToPath("A.Long.Path.to.the.Property", Dialect.POSTGRESQL, FieldFormat.SQL),
                "Path not constructed correctly");
    }

    @Test
    @DisplayName("nameToPath creates a nested SQLite SQL name")
    public void nameToPathSQLiteNestedSQL() {
        assertEquals("data->'A'->'Long'->'Path'->'to'->'the'->>'Property'",
                Field.nameToPath("A.Long.Path.to.the.Property", Dialect.SQLITE, FieldFormat.SQL),
                "Path not constructed correctly");
    }

    @Test
    @DisplayName("nameToPath creates a simple PostgreSQL JSON name")
    public void nameToPathPostgresSimpleJSON() {
        assertEquals("data->'Simple'", Field.nameToPath("Simple", Dialect.POSTGRESQL, FieldFormat.JSON),
                "Path not constructed correctly");
    }

    @Test
    @DisplayName("nameToPath creates a simple SQLite JSON name")
    public void nameToPathSQLiteSimpleJSON() {
        assertEquals("data->'Simple'", Field.nameToPath("Simple", Dialect.SQLITE, FieldFormat.JSON),
                "Path not constructed correctly");
    }

    @Test
    @DisplayName("nameToPath creates a nested PostgreSQL JSON name")
    public void nameToPathPostgresNestedJSON() {
        assertEquals("data#>'{A,Long,Path,to,the,Property}'",
                Field.nameToPath("A.Long.Path.to.the.Property", Dialect.POSTGRESQL, FieldFormat.JSON),
                "Path not constructed correctly");
    }

    @Test
    @DisplayName("nameToPath creates a nested SQLite JSON name")
    public void nameToPathSQLiteNestedJSON() {
        assertEquals("data->'A'->'Long'->'Path'->'to'->'the'->'Property'",
                Field.nameToPath("A.Long.Path.to.the.Property", Dialect.SQLITE, FieldFormat.JSON),
                "Path not constructed correctly");
    }
}
