package solutions.bitbadger.documents.common.query

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import solutions.bitbadger.documents.common.Configuration
import solutions.bitbadger.documents.common.Dialect
import solutions.bitbadger.documents.common.DocumentException
import solutions.bitbadger.documents.common.DocumentIndex
import kotlin.test.assertEquals

/**
 * Unit tests for the `Definition` object
 */
@DisplayName("Kotlin | Common | Query: Definition")
class DefinitionTest {

    /** Test table name */
    private val tbl = "test_table"

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    fun cleanUp() {
        Configuration.dialectValue = null
    }

    @Test
    @DisplayName("ensureTableFor generates correctly")
    fun ensureTableFor() =
        assertEquals(
            "CREATE TABLE IF NOT EXISTS my.table (data JSONB NOT NULL)",
            Definition.ensureTableFor("my.table", "JSONB"), "CREATE TABLE statement not constructed correctly"
        )

    @Test
    @DisplayName("ensureTable generates correctly (PostgreSQL)")
    fun ensureTablePostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals("CREATE TABLE IF NOT EXISTS $tbl (data JSONB NOT NULL)", Definition.ensureTable(tbl))
    }

    @Test
    @DisplayName("ensureTable generates correctly (SQLite)")
    fun ensureTableSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        assertEquals("CREATE TABLE IF NOT EXISTS $tbl (data TEXT NOT NULL)", Definition.ensureTable(tbl))
    }

    @Test
    @DisplayName("ensureTable fails when no dialect is set")
    fun ensureTableFailsUnknown() {
        assertThrows<DocumentException> { Definition.ensureTable(tbl) }
    }

    @Test
    @DisplayName("ensureKey generates correctly with schema")
    fun ensureKeyWithSchema() =
        assertEquals(
            "CREATE UNIQUE INDEX IF NOT EXISTS idx_table_key ON test.table ((data->>'id'))",
            Definition.ensureKey("test.table", Dialect.POSTGRESQL),
            "CREATE INDEX for key statement with schema not constructed correctly"
        )

    @Test
    @DisplayName("ensureKey generates correctly without schema")
    fun ensureKeyWithoutSchema() =
        assertEquals(
            "CREATE UNIQUE INDEX IF NOT EXISTS idx_${tbl}_key ON $tbl ((data->>'id'))",
            Definition.ensureKey(tbl, Dialect.SQLITE),
            "CREATE INDEX for key statement without schema not constructed correctly"
        )

    @Test
    @DisplayName("ensureIndexOn generates multiple fields and directions")
    fun ensureIndexOnMultipleFields() =
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_table_gibberish ON test.table ((data->>'taco'), (data->>'guac') DESC, (data->>'salsa') ASC)",
            Definition.ensureIndexOn(
                "test.table", "gibberish", listOf("taco", "guac DESC", "salsa ASC"),
                Dialect.POSTGRESQL
            ),
            "CREATE INDEX for multiple field statement not constructed correctly"
        )

    @Test
    @DisplayName("ensureIndexOn generates nested PostgreSQL field")
    fun ensureIndexOnNestedPostgres() =
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_${tbl}_nest ON $tbl ((data#>>'{a,b,c}'))",
            Definition.ensureIndexOn(tbl, "nest", listOf("a.b.c"), Dialect.POSTGRESQL),
            "CREATE INDEX for nested PostgreSQL field incorrect"
        )

    @Test
    @DisplayName("ensureIndexOn generates nested SQLite field")
    fun ensureIndexOnNestedSQLite() =
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_${tbl}_nest ON $tbl ((data->'a'->'b'->>'c'))",
            Definition.ensureIndexOn(tbl, "nest", listOf("a.b.c"), Dialect.SQLITE),
            "CREATE INDEX for nested SQLite field incorrect"
        )

    @Test
    @DisplayName("ensureDocumentIndexOn generates Full for PostgreSQL")
    fun ensureDocumentIndexOnFullPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_${tbl}_document ON $tbl USING GIN (data)",
            Definition.ensureDocumentIndexOn(tbl, DocumentIndex.FULL),
            "CREATE INDEX for full document index incorrect"
        )
    }

    @Test
    @DisplayName("ensureDocumentIndexOn generates Optimized for PostgreSQL")
    fun ensureDocumentIndexOnOptimizedPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_${tbl}_document ON $tbl USING GIN (data jsonb_path_ops)",
            Definition.ensureDocumentIndexOn(tbl, DocumentIndex.OPTIMIZED),
            "CREATE INDEX for optimized document index incorrect"
        )
    }

    @Test
    @DisplayName("ensureDocumentIndexOn fails for SQLite")
    fun ensureDocumentIndexOnFailsSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        assertThrows<DocumentException> {  Definition.ensureDocumentIndexOn(tbl, DocumentIndex.FULL) }
    }
}
