package solutions.bitbadger.documents.common.query

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import solutions.bitbadger.documents.common.AutoId
import solutions.bitbadger.documents.common.Configuration
import solutions.bitbadger.documents.common.Dialect
import solutions.bitbadger.documents.common.DocumentException
import kotlin.test.assertEquals
import kotlin.test.assertTrue

/**
 * Unit tests for the `Document` object
 */
@DisplayName("Kotlin | Common | Query: Document")
class DocumentTest {

    /** Test table name */
    private val tbl = "test_table"

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    fun cleanUp() {
        Configuration.dialectValue = null
    }

    @Test
    @DisplayName("insert generates no auto ID (PostgreSQL)")
    fun insertNoAutoPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals("INSERT INTO $tbl VALUES (:data)", Document.insert(tbl))
    }

    @Test
    @DisplayName("insert generates no auto ID (SQLite)")
    fun insertNoAutoSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        assertEquals("INSERT INTO $tbl VALUES (:data)", Document.insert(tbl))
    }

    @Test
    @DisplayName("insert generates auto number (PostgreSQL)")
    fun insertAutoNumberPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals(
            "INSERT INTO $tbl VALUES (:data::jsonb || ('{\"id\":' " +
                    "|| (SELECT COALESCE(MAX((data->>'id')::numeric), 0) + 1 FROM $tbl) || '}')::jsonb)",
            Document.insert(tbl, AutoId.NUMBER)
        )
    }

    @Test
    @DisplayName("insert generates auto number (SQLite)")
    fun insertAutoNumberSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        assertEquals(
            "INSERT INTO $tbl VALUES (json_set(:data, '$.id', " +
                    "(SELECT coalesce(max(data->>'id'), 0) + 1 FROM $tbl)))",
            Document.insert(tbl, AutoId.NUMBER)
        )
    }

    @Test
    @DisplayName("insert generates auto UUID (PostgreSQL)")
    fun insertAutoUUIDPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        val query = Document.insert(tbl, AutoId.UUID)
        assertTrue(
            query.startsWith("INSERT INTO $tbl VALUES (:data::jsonb || '{\"id\":\""),
            "Query start not correct (actual: $query)"
        )
        assertTrue(query.endsWith("\"}')"), "Query end not correct")
    }

    @Test
    @DisplayName("insert generates auto UUID (SQLite)")
    fun insertAutoUUIDSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        val query = Document.insert(tbl, AutoId.UUID)
        assertTrue(
            query.startsWith("INSERT INTO $tbl VALUES (json_set(:data, '$.id', '"),
            "Query start not correct (actual: $query)"
        )
        assertTrue(query.endsWith("'))"), "Query end not correct")
    }

    @Test
    @DisplayName("insert generates auto random string (PostgreSQL)")
    fun insertAutoRandomPostgres() {
        try {
            Configuration.dialectValue = Dialect.POSTGRESQL
            Configuration.idStringLength = 8
            val query = Document.insert(tbl, AutoId.RANDOM_STRING)
            assertTrue(
                query.startsWith("INSERT INTO $tbl VALUES (:data::jsonb || '{\"id\":\""),
                "Query start not correct (actual: $query)"
            )
            assertTrue(query.endsWith("\"}')"), "Query end not correct")
            assertEquals(
                8,
                query.replace("INSERT INTO $tbl VALUES (:data::jsonb || '{\"id\":\"", "").replace("\"}')", "").length,
                "Random string length incorrect"
            )
        } finally {
            Configuration.idStringLength = 16
        }
    }

    @Test
    @DisplayName("insert generates auto random string (SQLite)")
    fun insertAutoRandomSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        val query = Document.insert(tbl, AutoId.RANDOM_STRING)
        assertTrue(
            query.startsWith("INSERT INTO $tbl VALUES (json_set(:data, '$.id', '"),
            "Query start not correct (actual: $query)"
        )
        assertTrue(query.endsWith("'))"), "Query end not correct")
        assertEquals(
            Configuration.idStringLength,
            query.replace("INSERT INTO $tbl VALUES (json_set(:data, '$.id', '", "").replace("'))", "").length,
            "Random string length incorrect"
        )
    }

    @Test
    @DisplayName("insert fails when no dialect is set")
    fun insertFailsUnknown() {
        assertThrows<DocumentException> { Document.insert(tbl) }
    }

    @Test
    @DisplayName("save generates correctly")
    fun save() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals(
            "INSERT INTO $tbl VALUES (:data) ON CONFLICT ((data->>'id')) DO UPDATE SET data = EXCLUDED.data",
            Document.save(tbl), "INSERT ON CONFLICT UPDATE statement not constructed correctly"
        )
    }

    @Test
    @DisplayName("update generates successfully")
    fun update() =
        assertEquals("UPDATE $tbl SET data = :data", Document.update(tbl), "Update query not constructed correctly")
}
