package solutions.bitbadger.documents.common.query

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import solutions.bitbadger.documents.common.Configuration
import solutions.bitbadger.documents.common.Dialect
import solutions.bitbadger.documents.common.Field
import solutions.bitbadger.documents.common.FieldMatch
import kotlin.test.assertEquals

/**
 * Unit tests for the top-level query functions
 */
@DisplayName("Kotlin | Common | Query")
class QueryTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    fun cleanUp() {
        Configuration.dialectValue = null
    }

    @Test
    @DisplayName("statementWhere generates correctly")
    fun statementWhere() =
        assertEquals("x WHERE y", statementWhere("x", "y"), "Statements not combined correctly")

    @Test
    @DisplayName("byId generates a numeric ID query (PostgreSQL)")
    fun byIdNumericPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals("test WHERE (data->>'id')::numeric = :id", byId("test", 9))
    }

    @Test
    @DisplayName("byId generates an alphanumeric ID query (PostgreSQL)")
    fun byIdAlphaPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals("unit WHERE data->>'id' = :id", byId("unit", "18"))
    }

    @Test
    @DisplayName("byId generates ID query (SQLite)")
    fun byIdSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        assertEquals("yo WHERE data->>'id' = :id", byId("yo", 27))
    }

    @Test
    @DisplayName("byFields generates default field query (PostgreSQL)")
    fun byFieldsMultipleDefaultPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals(
            "this WHERE data->>'a' = :the_a AND (data->>'b')::numeric = :b_value",
            byFields("this", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")))
        )
    }

    @Test
    @DisplayName("byFields generates default field query (SQLite)")
    fun byFieldsMultipleDefaultSQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        assertEquals(
            "this WHERE data->>'a' = :the_a AND data->>'b' = :b_value",
            byFields("this", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")))
        )
    }

    @Test
    @DisplayName("byFields generates ANY field query (PostgreSQL)")
    fun byFieldsMultipleAnyPostgres() {
        Configuration.dialectValue = Dialect.POSTGRESQL
        assertEquals(
            "that WHERE data->>'a' = :the_a OR (data->>'b')::numeric = :b_value",
            byFields(
                "that", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")),
                FieldMatch.ANY
            )
        )
    }

    @Test
    @DisplayName("byFields generates ANY field query (SQLite)")
    fun byFieldsMultipleAnySQLite() {
        Configuration.dialectValue = Dialect.SQLITE
        assertEquals(
            "that WHERE data->>'a' = :the_a OR data->>'b' = :b_value",
            byFields(
                "that", listOf(Field.equal("a", "", ":the_a"), Field.equal("b", 0, ":b_value")),
                FieldMatch.ANY
            )
        )
    }

    @Test
    @DisplayName("orderBy generates for no fields")
    fun orderByNone() {
        assertEquals("", orderBy(listOf(), Dialect.POSTGRESQL), "ORDER BY should have been blank (PostgreSQL)")
        assertEquals("", orderBy(listOf(), Dialect.SQLITE), "ORDER BY should have been blank (SQLite)")
    }

    @Test
    @DisplayName("orderBy generates single, no direction for PostgreSQL")
    fun orderBySinglePostgres() =
        assertEquals(
            " ORDER BY data->>'TestField'",
            orderBy(listOf(Field.named("TestField")), Dialect.POSTGRESQL), "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates single, no direction for SQLite")
    fun orderBySingleSQLite() =
        assertEquals(
            " ORDER BY data->>'TestField'", orderBy(listOf(Field.named("TestField")), Dialect.SQLITE),
            "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates multiple with direction for PostgreSQL")
    fun orderByMultiplePostgres() =
        assertEquals(
            " ORDER BY data#>>'{Nested,Test,Field}' DESC, data->>'AnotherField', data->>'It' DESC",
            orderBy(
                listOf(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")),
                Dialect.POSTGRESQL
            ),
            "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates multiple with direction for SQLite")
    fun orderByMultipleSQLite() =
        assertEquals(
            " ORDER BY data->'Nested'->'Test'->>'Field' DESC, data->>'AnotherField', data->>'It' DESC",
            orderBy(
                listOf(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")),
                Dialect.SQLITE
            ),
            "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates numeric ordering PostgreSQL")
    fun orderByNumericPostgres() =
        assertEquals(
            " ORDER BY (data->>'Test')::numeric",
            orderBy(listOf(Field.named("n:Test")), Dialect.POSTGRESQL), "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates numeric ordering for SQLite")
    fun orderByNumericSQLite() =
        assertEquals(
            " ORDER BY data->>'Test'", orderBy(listOf(Field.named("n:Test")), Dialect.SQLITE),
            "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates case-insensitive ordering for PostgreSQL")
    fun orderByCIPostgres() =
        assertEquals(
            " ORDER BY LOWER(data#>>'{Test,Field}') DESC NULLS FIRST",
            orderBy(listOf(Field.named("i:Test.Field DESC NULLS FIRST")), Dialect.POSTGRESQL),
            "ORDER BY not constructed correctly"
        )

    @Test
    @DisplayName("orderBy generates case-insensitive ordering for SQLite")
    fun orderByCISQLite() =
        assertEquals(
            " ORDER BY data->'Test'->>'Field' COLLATE NOCASE ASC NULLS LAST",
            orderBy(listOf(Field.named("i:Test.Field ASC NULLS LAST")), Dialect.SQLITE),
            "ORDER BY not constructed correctly"
        )
}
