package solutions.bitbadger.documents.java

import solutions.bitbadger.documents.common.*
import solutions.bitbadger.documents.common.query.Delete
import java.sql.Connection

/**
 * Functions to delete documents
 */
object Delete {

    /**
     * Delete a document by its ID
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param docId The ID of the document to be deleted
     * @param conn The connection on which the deletion should be executed
     */
    @JvmStatic
    fun <TKey> byId(tableName: String, docId: TKey, conn: Connection) =
        conn.customNonQuery(
            Delete.byId(tableName, docId),
            Parameters.addFields(listOf(Field.equal(Configuration.idField, docId, ":id")))
        )

    /**
     * Delete a document by its ID
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param docId The ID of the document to be deleted
     */
    @JvmStatic
    fun <TKey> byId(tableName: String, docId: TKey) =
        Configuration.dbConn().use { byId(tableName, docId, it) }

    /**
     * Delete documents using a field comparison
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @param conn The connection on which the deletion should be executed
     */
    @JvmStatic
    @JvmOverloads
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null, conn: Connection) {
        val named = Parameters.nameFields(fields)
        conn.customNonQuery(Delete.byFields(tableName, named, howMatched), Parameters.addFields(named))
    }

    /**
     * Delete documents using a field comparison
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     */
    @JvmStatic
    @JvmOverloads
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null) =
        Configuration.dbConn().use { byFields(tableName, fields, howMatched, it) }

    /**
     * Delete documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param criteria The object for which JSON containment should be checked
     * @param conn The connection on which the deletion should be executed
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> byContains(tableName: String, criteria: TContains, conn: Connection) =
        conn.customNonQuery(Delete.byContains(tableName), listOf(Parameters.json(":criteria", criteria)))

    /**
     * Delete documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param criteria The object for which JSON containment should be checked
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> byContains(tableName: String, criteria: TContains) =
        Configuration.dbConn().use { byContains(tableName, criteria, it) }

    /**
     * Delete documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param path The JSON path comparison to match
     * @param conn The connection on which the deletion should be executed
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byJsonPath(tableName: String, path: String, conn: Connection) =
        conn.customNonQuery(Delete.byJsonPath(tableName), listOf(Parameter(":path", ParameterType.STRING, path)))

    /**
     * Delete documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param path The JSON path comparison to match
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byJsonPath(tableName: String, path: String) =
        Configuration.dbConn().use { byJsonPath(tableName, path, it) }
}
