package solutions.bitbadger.documents.java

import solutions.bitbadger.documents.common.*
import solutions.bitbadger.documents.common.query.Find
import solutions.bitbadger.documents.common.query.orderBy
import java.sql.Connection

/**
 * Functions to find and retrieve documents
 */
object Find {

    /**
     * Retrieve all documents in the given table, ordering results by the optional given fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents from the given table
     */
    @JvmStatic
    fun <TDoc> all(tableName: String, clazz: Class<TDoc>, orderBy: Collection<Field<*>>? = null, conn: Connection) =
        conn.customList(Find.all(tableName) + (orderBy?.let(::orderBy) ?: ""), listOf(), clazz, Results::fromData)

    /**
     * Retrieve all documents in the given table
     *
     * @param tableName The table from which documents should be retrieved
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents from the given table
     */
    @JvmStatic
    @JvmOverloads
    fun <TDoc> all(tableName: String, clazz: Class<TDoc>, orderBy: Collection<Field<*>>? = null) =
        Configuration.dbConn().use { all(tableName, clazz, orderBy, it) }

    /**
     * Retrieve all documents in the given table
     *
     * @param tableName The table from which documents should be retrieved
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents from the given table
     */
    @JvmStatic
    fun <TDoc> all(tableName: String, clazz: Class<TDoc>, conn: Connection) =
        all(tableName, clazz, null, conn)

    /**
     * Retrieve a document by its ID
     *
     * @param tableName The table from which the document should be retrieved
     * @param docId The ID of the document to retrieve
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return The document if it is found, `null` otherwise
     */
    @JvmStatic
    fun <TKey, TDoc> byId(tableName: String, docId: TKey, clazz: Class<TDoc>, conn: Connection) =
        conn.customSingle(
            Find.byId(tableName, docId),
            Parameters.addFields(listOf(Field.equal(Configuration.idField, docId, ":id"))),
            clazz,
            Results::fromData
        )

    /**
     * Retrieve a document by its ID
     *
     * @param tableName The table from which the document should be retrieved
     * @param docId The ID of the document to retrieve
     * @param clazz The class of the document to be returned
     * @return The document if it is found, `null` otherwise
     */
    @JvmStatic
    fun <TKey, TDoc> byId(tableName: String, docId: TKey, clazz: Class<TDoc>) =
        Configuration.dbConn().use { byId(tableName, docId, clazz, it) }

    /**
     * Retrieve documents using a field comparison, ordering results by the given fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the field comparison
     */
    @JvmStatic
    fun <TDoc> byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ): List<TDoc> {
        val named = Parameters.nameFields(fields)
        return conn.customList(
            Find.byFields(tableName, named, howMatched) + (orderBy?.let(::orderBy) ?: ""),
            Parameters.addFields(named),
            clazz,
            Results::fromData
        )
    }

    /**
     * Retrieve documents using a field comparison, ordering results by the given fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents matching the field comparison
     */
    @JvmStatic
    @JvmOverloads
    fun <TDoc> byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null
    ) =
        Configuration.dbConn().use { byFields(tableName, fields, clazz, howMatched, orderBy, it) }

    /**
     * Retrieve documents using a field comparison
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the field comparison
     */
    @JvmStatic
    fun <TDoc> byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        conn: Connection
    ) =
        byFields(tableName, fields, clazz, howMatched, null, conn)

    /**
     * Retrieve documents using a JSON containment query, ordering results by the given fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the JSON containment query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc, TContains> byContains(
        tableName: String,
        criteria: TContains,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) =
        conn.customList(
            Find.byContains(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(Parameters.json(":criteria", criteria)),
            clazz,
            Results::fromData
        )

    /**
     * Retrieve documents using a JSON containment query, ordering results by the given fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents matching the JSON containment query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TDoc, TContains> byContains(
        tableName: String,
        criteria: TContains,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null
    ) =
        Configuration.dbConn().use { byContains(tableName, criteria, clazz, orderBy, it) }

    /**
     * Retrieve documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the JSON containment query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc, TContains> byContains(tableName: String, criteria: TContains, clazz: Class<TDoc>, conn: Connection) =
        byContains(tableName, criteria, clazz, null, conn)

    /**
     * Retrieve documents using a JSON Path match query, ordering results by the given fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the JSON Path match query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> byJsonPath(
        tableName: String,
        path: String,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) =
        conn.customList(
            Find.byJsonPath(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(Parameter(":path", ParameterType.STRING, path)),
            clazz,
            Results::fromData
        )

    /**
     * Retrieve documents using a JSON Path match query, ordering results by the given fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents matching the JSON Path match query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TDoc> byJsonPath(tableName: String, path: String, clazz: Class<TDoc>, orderBy: Collection<Field<*>>? = null) =
        Configuration.dbConn().use { byJsonPath(tableName, path, clazz, orderBy, it) }

    /**
     * Retrieve documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the JSON Path match query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> byJsonPath(tableName: String, path: String, clazz: Class<TDoc>, conn: Connection) =
        byJsonPath(tableName, path, clazz, null, conn)

    /**
     * Retrieve the first document using a field comparison and optional ordering fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return The first document matching the field comparison, or `null` if no matches are found
     */
    @JvmStatic
    fun <TDoc> firstByFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ): TDoc? {
        val named = Parameters.nameFields(fields)
        return conn.customSingle(
            Find.byFields(tableName, named, howMatched) + (orderBy?.let(::orderBy) ?: ""),
            Parameters.addFields(named),
            clazz,
            Results::fromData
        )
    }

    /**
     * Retrieve the first document using a field comparison and optional ordering fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return The first document matching the field comparison, or `null` if no matches are found
     */
    @JvmStatic
    @JvmOverloads
    fun <TDoc> firstByFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null
    ) =
        Configuration.dbConn().use { firstByFields(tableName, fields, clazz, howMatched, orderBy, it) }

    /**
     * Retrieve the first document using a field comparison
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param conn The connection over which documents should be retrieved
     * @return The first document matching the field comparison, or `null` if no matches are found
     */
    @JvmStatic
    fun <TDoc> firstByFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        conn: Connection
    ) =
        firstByFields(tableName, fields, clazz, howMatched, null, conn)

    /**
     * Retrieve the first document using a JSON containment query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return The first document matching the JSON containment query, or `null` if no matches are found
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc, TContains> firstByContains(
        tableName: String,
        criteria: TContains,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) =
        conn.customSingle(
            Find.byContains(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(Parameters.json(":criteria", criteria)),
            clazz,
            Results::fromData
        )

    /**
     * Retrieve the first document using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return The first document matching the JSON containment query, or `null` if no matches are found
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc, TContains> firstByContains(
        tableName: String,
        criteria: TContains,
        clazz: Class<TDoc>,
        conn: Connection
    ) =
        firstByContains(tableName, criteria, clazz, null, conn)

    /**
     * Retrieve the first document using a JSON containment query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return The first document matching the JSON containment query, or `null` if no matches are found
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TDoc, TContains> firstByContains(
        tableName: String,
        criteria: TContains,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null
    ) =
        Configuration.dbConn().use { firstByContains(tableName, criteria, clazz, orderBy, it) }

    /**
     * Retrieve the first document using a JSON Path match query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return The first document matching the JSON Path match query, or `null` if no matches are found
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> firstByJsonPath(
        tableName: String,
        path: String,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) =
        conn.customSingle(
            Find.byJsonPath(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(Parameter(":path", ParameterType.STRING, path)),
            clazz,
            Results::fromData
        )

    /**
     * Retrieve the first document using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return The first document matching the JSON Path match query, or `null` if no matches are found
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> firstByJsonPath(tableName: String, path: String, clazz: Class<TDoc>, conn: Connection) =
        firstByJsonPath(tableName, path, clazz, null, conn)

    /**
     * Retrieve the first document using a JSON Path match query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return The first document matching the JSON Path match query, or `null` if no matches are found
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TDoc> firstByJsonPath(
        tableName: String,
        path: String,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null
    ) =
        Configuration.dbConn().use { firstByJsonPath(tableName, path, clazz, orderBy, it) }
}
