package solutions.bitbadger.documents.java.java;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import solutions.bitbadger.documents.common.*;
import solutions.bitbadger.documents.java.Parameters;

import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

/**
 * Unit tests for the `Parameters` object
 */
@DisplayName("Java | Java | Parameters")
final public class ParametersTest {

    /**
     * Reset the dialect
     */
    @AfterEach
    public void cleanUp() {
        Configuration.setConnectionString(null);
    }

    @Test
    @DisplayName("nameFields works with no changes")
    public void nameFieldsNoChange() {
        List<Field<?>> fields = List.of(Field.equal("a", "", ":test"), Field.exists("q"), Field.equal("b", "", ":me"));
        Field<?>[] named = Parameters.nameFields(fields).toArray(new Field<?>[] { });
        assertEquals(fields.size(), named.length, "There should have been 3 fields in the list");
        assertSame(fields.get(0), named[0], "The first field should be the same");
        assertSame(fields.get(1), named[1], "The second field should be the same");
        assertSame(fields.get(2), named[2], "The third field should be the same");
    }

    @Test
    @DisplayName("nameFields works when changing fields")
    public void nameFieldsChange() {
        List<Field<?>> fields = List.of(
                Field.equal("a", ""), Field.equal("e", "", ":hi"), Field.equal("b", ""), Field.notExists("z"));
        Field<?>[] named = Parameters.nameFields(fields).toArray(new Field<?>[] { });
        assertEquals(fields.size(), named.length, "There should have been 4 fields in the list");
        assertNotSame(fields.get(0), named[0], "The first field should not be the same");
        assertEquals(":field0", named[0].getParameterName(), "First parameter name incorrect");
        assertSame(fields.get(1), named[1], "The second field should be the same");
        assertNotSame(fields.get(2), named[2], "The third field should not be the same");
        assertEquals(":field1", named[2].getParameterName(), "Third parameter name incorrect");
        assertSame(fields.get(3), named[3], "The fourth field should be the same");
    }

    @Test
    @DisplayName("replaceNamesInQuery replaces successfully")
    public void replaceNamesInQuery() {
        List<Parameter<?>> parameters = List.of(new Parameter<>(":data", ParameterType.JSON, "{}"),
                new Parameter<>(":data_ext", ParameterType.STRING, ""));
        String query =
                "SELECT data, data_ext FROM tbl WHERE data = :data AND data_ext = :data_ext AND more_data = :data";
        assertEquals("SELECT data, data_ext FROM tbl WHERE data = ? AND data_ext = ? AND more_data = ?",
                Parameters.replaceNamesInQuery(query, parameters), "Parameters not replaced correctly");
    }

    @Test
    @DisplayName("fieldNames generates a single parameter (PostgreSQL)")
    public void fieldNamesSinglePostgres() throws DocumentException {
        Configuration.setConnectionString(":postgresql:");
        Parameter<?>[] nameParams = Parameters.fieldNames(List.of("test")).toArray(new Parameter<?>[]{});
        assertEquals(1, nameParams.length, "There should be one name parameter");
        assertEquals(":name", nameParams[0].getName(), "The parameter name is incorrect");
        assertEquals(ParameterType.STRING, nameParams[0].getType(), "The parameter type is incorrect");
        assertEquals("{test}", nameParams[0].getValue(), "The parameter value is incorrect");
    }

    @Test
    @DisplayName("fieldNames generates multiple parameters (PostgreSQL)")
    public void fieldNamesMultiplePostgres() throws DocumentException {
        Configuration.setConnectionString(":postgresql:");
        Parameter<?>[] nameParams = Parameters.fieldNames(List.of("test", "this", "today"))
                .toArray(new Parameter<?>[]{});
        assertEquals(1, nameParams.length, "There should be one name parameter");
        assertEquals(":name", nameParams[0].getName(), "The parameter name is incorrect");
        assertEquals(ParameterType.STRING, nameParams[0].getType(), "The parameter type is incorrect");
        assertEquals("{test,this,today}", nameParams[0].getValue(), "The parameter value is incorrect");
    }

    @Test
    @DisplayName("fieldNames generates a single parameter (SQLite)")
    public void fieldNamesSingleSQLite() throws DocumentException {
        Configuration.setConnectionString(":sqlite:");
        Parameter<?>[] nameParams = Parameters.fieldNames(List.of("test")).toArray(new Parameter<?>[]{});
        assertEquals(1, nameParams.length, "There should be one name parameter");
        assertEquals(":name0", nameParams[0].getName(), "The parameter name is incorrect");
        assertEquals(ParameterType.STRING, nameParams[0].getType(), "The parameter type is incorrect");
        assertEquals("test", nameParams[0].getValue(), "The parameter value is incorrect");
    }

    @Test
    @DisplayName("fieldNames generates multiple parameters (SQLite)")
    public void fieldNamesMultipleSQLite() throws DocumentException {
        Configuration.setConnectionString(":sqlite:");
        Parameter<?>[] nameParams = Parameters.fieldNames(List.of("test", "this", "today"))
                .toArray(new Parameter<?>[]{});
        assertEquals(3, nameParams.length, "There should be one name parameter");
        assertEquals(":name0", nameParams[0].getName(), "The first parameter name is incorrect");
        assertEquals(ParameterType.STRING, nameParams[0].getType(), "The first parameter type is incorrect");
        assertEquals("test", nameParams[0].getValue(), "The first parameter value is incorrect");
        assertEquals(":name1", nameParams[1].getName(), "The second parameter name is incorrect");
        assertEquals(ParameterType.STRING, nameParams[1].getType(), "The second parameter type is incorrect");
        assertEquals("this", nameParams[1].getValue(), "The second parameter value is incorrect");
        assertEquals(":name2", nameParams[2].getName(), "The third parameter name is incorrect");
        assertEquals(ParameterType.STRING, nameParams[2].getType(), "The third parameter type is incorrect");
        assertEquals("today", nameParams[2].getValue(), "The third parameter value is incorrect");
    }

    @Test
    @DisplayName("fieldNames fails if dialect not set")
    public void fieldNamesFails() {
        assertThrows(DocumentException.class, () -> Parameters.fieldNames(List.of()));
    }
}
