package solutions.bitbadger.documents.java.integration.common

import solutions.bitbadger.documents.common.Field
import solutions.bitbadger.documents.java.*
import solutions.bitbadger.documents.java.integration.JsonDocument
import solutions.bitbadger.documents.java.integration.TEST_TABLE
import solutions.bitbadger.documents.java.integration.ThrowawayDatabase
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertNotNull
import kotlin.test.assertTrue

/**
 * Integration tests for the `Patch` object
 */
object Patch {

    fun byIdMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        db.conn.patchById(TEST_TABLE, "one", mapOf("numValue" to 44))
        val doc = db.conn.findById(TEST_TABLE, "one", JsonDocument::class.java)
        assertNotNull(doc, "There should have been a document returned")
        assertEquals("one", doc.id, "An incorrect document was returned")
        assertEquals(44, doc.numValue, "The document was not patched")
    }

    fun byIdNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertFalse("Document with ID \"forty-seven\" should not exist") {
            db.conn.existsById(TEST_TABLE, "forty-seven")
        }
        db.conn.patchById(TEST_TABLE, "forty-seven", mapOf("foo" to "green")) // no exception = pass
    }

    fun byFieldsMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        db.conn.patchByFields(TEST_TABLE, listOf(Field.equal("value", "purple")), mapOf("numValue" to 77))
        assertEquals(
            2,
            db.conn.countByFields(TEST_TABLE, listOf(Field.equal("numValue", 77))),
            "There should have been 2 documents with numeric value 77"
        )
    }

    fun byFieldsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val fields = listOf(Field.equal("value", "burgundy"))
        assertFalse("There should be no documents with value of \"burgundy\"") {
            db.conn.existsByFields(TEST_TABLE, fields)
        }
        db.conn.patchByFields(TEST_TABLE, fields, mapOf("foo" to "green")) // no exception = pass
    }

    fun byContainsMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val contains = mapOf("value" to "another")
        db.conn.patchByContains(TEST_TABLE, contains, mapOf("numValue" to 12))
        val doc = db.conn.findFirstByContains(TEST_TABLE, contains, JsonDocument::class.java)
        assertNotNull(doc, "There should have been a document returned")
        assertEquals("two", doc.id, "The incorrect document was returned")
        assertEquals(12, doc.numValue, "The document was not updated")
    }

    fun byContainsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val contains = mapOf("value" to "updated")
        assertFalse("There should be no matching documents") { db.conn.existsByContains(TEST_TABLE, contains) }
        db.conn.patchByContains(TEST_TABLE, contains, mapOf("sub.foo" to "green")) // no exception = pass
    }

    fun byJsonPathMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val path = "$.numValue ? (@ > 10)"
        db.conn.patchByJsonPath(TEST_TABLE, path, mapOf("value" to "blue"))
        val docs = db.conn.findByJsonPath(TEST_TABLE, path, JsonDocument::class.java)
        assertEquals(2, docs.size, "There should have been two documents returned")
        docs.forEach {
            assertTrue(listOf("four", "five").contains(it.id), "An incorrect document was returned (${it.id})")
            assertEquals("blue", it.value, "The value for ID ${it.id} was incorrect")
        }
    }

    fun byJsonPathNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val path = "$.numValue ? (@ > 100)"
        assertFalse("There should be no documents with numeric values over 100") {
            db.conn.existsByJsonPath(TEST_TABLE, path)
        }
        db.conn.patchByJsonPath(TEST_TABLE, path, mapOf("value" to "blue")) // no exception = pass
    }
}
