package solutions.bitbadger.documents.kotlin

import solutions.bitbadger.documents.common.Configuration
import solutions.bitbadger.documents.common.Parameter
import solutions.bitbadger.documents.java.Custom
import java.sql.Connection
import java.sql.ResultSet

/**
 * Custom query execution functions
 */
object Custom {

    /**
     * Execute a query that returns a list of results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param clazz The class of the document to be returned
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function between the document and the domain item
     * @return A list of results for the given query
     */
    inline fun <reified TDoc> list(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        conn: Connection,
        noinline mapFunc: (ResultSet, Class<TDoc>) -> TDoc
    ) = Custom.list(query, parameters, TDoc::class.java, conn, mapFunc)

    /**
     * Execute a query that returns a list of results (creates connection)
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param mapFunc The mapping function between the document and the domain item
     * @return A list of results for the given query
     */
    inline fun <reified TDoc> list(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        noinline mapFunc: (ResultSet, Class<TDoc>) -> TDoc
    ) = Configuration.dbConn().use { list(query, parameters, it, mapFunc) }

    /**
     * Execute a query that returns one or no results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function between the document and the domain item
     * @return The document if one matches the query, `null` otherwise
     */
    inline fun <reified TDoc> single(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        conn: Connection,
        noinline mapFunc: (ResultSet, Class<TDoc>) -> TDoc
    ) = Custom.single(query, parameters, TDoc::class.java, conn, mapFunc)

    /**
     * Execute a query that returns one or no results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param mapFunc The mapping function between the document and the domain item
     * @return The document if one matches the query, `null` otherwise
     */
    inline fun <reified TDoc> single(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        noinline mapFunc: (ResultSet, Class<TDoc>) -> TDoc
    ) = Configuration.dbConn().use { single(query, parameters, it, mapFunc) }

    /**
     * Execute a query that returns no results
     *
     * @param query The query to retrieve the results
     * @param conn The connection over which the query should be executed
     * @param parameters Parameters to use for the query
     */
    fun nonQuery(query: String, parameters: Collection<Parameter<*>> = listOf(), conn: Connection) =
        Custom.nonQuery(query, parameters, conn)

    /**
     * Execute a query that returns no results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     */
    fun nonQuery(query: String, parameters: Collection<Parameter<*>> = listOf()) =
        Configuration.dbConn().use { nonQuery(query, parameters, it) }

    /**
     * Execute a query that returns a scalar result
     *
     * @param query The query to retrieve the result
     * @param parameters Parameters to use for the query
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function between the document and the domain item
     * @return The scalar value from the query
     */
    inline fun <reified T : Any> scalar(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        conn: Connection,
        noinline mapFunc: (ResultSet, Class<T>) -> T
    ) = Custom.scalar(query, parameters, T::class.java, conn, mapFunc)

    /**
     * Execute a query that returns a scalar result
     *
     * @param query The query to retrieve the result
     * @param parameters Parameters to use for the query
     * @param mapFunc The mapping function between the document and the domain item
     * @return The scalar value from the query
     */
    inline fun <reified T : Any> scalar(
        query: String, parameters: Collection<Parameter<*>> = listOf(), noinline mapFunc: (ResultSet, Class<T>) -> T
    ) = Configuration.dbConn().use { scalar(query, parameters, it, mapFunc) }
}
