import solutions.bitbadger.documents.common.*
import java.sql.Connection

/**
 * Functions to remove fields from documents
 */
object RemoveFields {

    /**
     * Translate field paths to JSON paths for SQLite queries
     *
     * @param parameters The parameters for the specified fields
     * @return The parameters for the specified fields, translated if used for SQLite
     */
    private fun translatePath(parameters: MutableCollection<Parameter<*>>): MutableCollection<Parameter<*>> {
        val dialect = Configuration.dialect("remove fields")
        return when (dialect) {
            Dialect.POSTGRESQL -> parameters
            Dialect.SQLITE     -> parameters.map { Parameter(it.name, it.type, "$.${it.value}") }.toMutableList()
        }
    }

    /**
     * Remove fields from a document by its ID
     *
     * @param tableName The name of the table in which the document's fields should be removed
     * @param docId The ID of the document to have fields removed
     * @param toRemove The names of the fields to be removed
     * @param conn The connection on which the update should be executed
     */
    fun <TKey> byId(tableName: String, docId: TKey, toRemove: Collection<String>, conn: Connection) {
        val nameParams = Parameters.fieldNames(toRemove)
        conn.customNonQuery(
            RemoveFields.byId(tableName, nameParams, docId),
            Parameters.addFields(
                listOf(Field.equal(Configuration.idField, docId, ":id")),
                translatePath(nameParams)
            )
        )
    }

    /**
     * Remove fields from a document by its ID
     *
     * @param tableName The name of the table in which the document's fields should be removed
     * @param docId The ID of the document to have fields removed
     * @param toRemove The names of the fields to be removed
     */
    fun <TKey> byId(tableName: String, docId: TKey, toRemove: Collection<String>) =
        Configuration.dbConn().use { byId(tableName, docId, toRemove, it) }

    /**
     * Remove fields from documents using a field comparison
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param fields The fields which should be compared
     * @param toRemove The names of the fields to be removed
     * @param howMatched How the fields should be matched
     * @param conn The connection on which the update should be executed
     */
    fun byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        toRemove: Collection<String>,
        howMatched: FieldMatch? = null,
        conn: Connection
    ) {
        val named = Parameters.nameFields(fields)
        val nameParams = Parameters.fieldNames(toRemove)
        conn.customNonQuery(
            RemoveFields.byFields(tableName, nameParams, named, howMatched),
            Parameters.addFields(named, translatePath(nameParams))
        )
    }

    /**
     * Remove fields from documents using a field comparison
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param fields The fields which should be compared
     * @param toRemove The names of the fields to be removed
     * @param howMatched How the fields should be matched
     */
    fun byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        toRemove: Collection<String>,
        howMatched: FieldMatch? = null
    ) =
        Configuration.dbConn().use { byFields(tableName, fields, toRemove, howMatched, it) }

    /**
     * Remove fields from documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param criteria The object against which JSON containment should be checked
     * @param toRemove The names of the fields to be removed
     * @param conn The connection on which the update should be executed
     * @throws DocumentException If called on a SQLite connection
     */
    inline fun <reified TContains> byContains(
        tableName: String,
        criteria: TContains,
        toRemove: Collection<String>,
        conn: Connection
    ) {
        val nameParams = Parameters.fieldNames(toRemove)
        conn.customNonQuery(
            RemoveFields.byContains(tableName, nameParams),
            listOf(Parameters.json(":criteria", criteria), *nameParams.toTypedArray())
        )
    }

    /**
     * Remove fields from documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param criteria The object against which JSON containment should be checked
     * @param toRemove The names of the fields to be removed
     * @throws DocumentException If called on a SQLite connection
     */
    inline fun <reified TContains> byContains(tableName: String, criteria: TContains, toRemove: Collection<String>) =
        Configuration.dbConn().use { byContains(tableName, criteria, toRemove, it) }

    /**
     * Remove fields from documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param path The JSON path comparison to match
     * @param toRemove The names of the fields to be removed
     * @param conn The connection on which the update should be executed
     * @throws DocumentException If called on a SQLite connection
     */
    fun byJsonPath(tableName: String, path: String, toRemove: Collection<String>, conn: Connection) {
        val nameParams = Parameters.fieldNames(toRemove)
        conn.customNonQuery(
            RemoveFields.byJsonPath(tableName, nameParams),
            listOf(Parameter(":path", ParameterType.STRING, path), *nameParams.toTypedArray())
        )
    }

    /**
     * Remove fields from documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param path The JSON path comparison to match
     * @param toRemove The names of the fields to be removed
     * @throws DocumentException If called on a SQLite connection
     */
    fun byJsonPath(tableName: String, path: String, toRemove: Collection<String>) =
        Configuration.dbConn().use { byJsonPath(tableName, path, toRemove, it) }
}
