package solutions.bitbadger.documents.kotlin

import solutions.bitbadger.documents.common.Configuration
import solutions.bitbadger.documents.common.Dialect
import solutions.bitbadger.documents.common.DocumentException
import solutions.bitbadger.documents.java.Results
import java.sql.PreparedStatement
import java.sql.ResultSet
import java.sql.SQLException

/**
 * Helper functions for handling results
 */
object Results {

    /**
     * Create a domain item from a document, specifying the field in which the document is found
     *
     * @param field The field name containing the JSON document
     * @param rs A `ResultSet` set to the row with the document to be constructed
     * @return The constructed domain item
     */
    inline fun <reified TDoc> fromDocument(field: String): (ResultSet, Class<TDoc>) -> TDoc =
        { rs, _ -> Results.fromDocument(field, rs, TDoc::class.java) }

    /**
     * Create a domain item from a document
     *
     * @param rs A `ResultSet` set to the row with the document to be constructed<
     * @param clazz The class of the document to be returned
     * @return The constructed domain item
     */
    inline fun <reified TDoc> fromData(rs: ResultSet, clazz: Class<TDoc> = TDoc::class.java) =
        Results.fromDocument("data", rs, TDoc::class.java)

    /**
     * Create a list of items for the results of the given command, using the specified mapping function
     *
     * @param stmt The prepared statement to execute
     * @param mapFunc The mapping function from data reader to domain class instance
     * @return A list of items from the query's result
     * @throws DocumentException If there is a problem executing the query
     */
    inline fun <reified TDoc : Any> toCustomList(stmt: PreparedStatement, mapFunc: (ResultSet) -> TDoc) =
        try {
            stmt.executeQuery().use {
                val results = mutableListOf<TDoc>()
                while (it.next()) {
                    results.add(mapFunc(it))
                }
                results.toList()
            }
        } catch (ex: SQLException) {
            throw DocumentException("Error retrieving documents from query: ${ex.message}", ex)
        }

    /**
     * Extract a count from the first column
     *
     * @param rs A `ResultSet` set to the row with the count to retrieve
     * @return The count from the row
     */
    fun toCount(rs: ResultSet, clazz: Class<Long> = Long::class.java) =
        when (Configuration.dialect()) {
            Dialect.POSTGRESQL -> rs.getInt("it").toLong()
            Dialect.SQLITE -> rs.getLong("it")
        }

    /**
     * Extract a true/false value from the first column
     *
     * @param rs A `ResultSet` set to the row with the true/false value to retrieve
     * @return The true/false value from the row
     */
    fun toExists(rs: ResultSet, clazz: Class<Boolean> = Boolean::class.java) =
        when (Configuration.dialect()) {
            Dialect.POSTGRESQL -> rs.getBoolean("it")
            Dialect.SQLITE -> toCount(rs) > 0L
        }
}
